import layout as cpp
import numpy as np

__all__ = ['layout','layout_convergent','layout_sparse','mds_direct','draw_svg','draw_png']

def layout(I, J, V=None, t_max=15, eps=.1):
    """takes a list of indices I and J
    and returns a n-by-2 matrix of positions X with minimized stress."""

    # initialize positions
    X = random_init(I, J)
    
    if V is None:
        cpp.layout_unweighted(X, I, J, t_max, eps)
    else:
        cpp.layout_weighted(X, I, J, V, t_max, eps)
    return X

def layout_convergent(I, J, V=None, t_max=30, eps=.1, delta=.03, t_maxmax=200):
    """takes a list of indices I and J
    and returns a n-by-2 matrix of positions X with minimized stress
    at a guaranteed stationary point."""

    # initialize positions
    X = random_init(I, J)

    if V is None:
        cpp.layout_unweighted_convergent(X, I, J, t_max, eps, delta, t_maxmax)
    else:
        cpp.layout_weighted_convergent(X, I, J, V, t_max, eps, delta, t_maxmax)

    return X

def layout_sparse(I, J, npivots, V=None, t_max=30, eps=.01):
    """takes a list of indices I and J
    and returns a n-by-2 matrix of positions X with minimized stress
    using the sparse approximation of Ortmann et al. (2017)"""

    # initialize positions
    X = random_init(I, J)

    if (npivots > X.shape[0]):
        raise "number of pivots exceeds number of vertices"

    if V is None:
        cpp.layout_sparse_unweighted(X, I, J, npivots, t_max, eps);
    else:
        cpp.layout_sparse_weighted(X, I, J, V, npivots, t_max, eps);

    return X


def mds_direct(n, d, w, etas=None):
    """takes nC2 vectors d and w with a vector of step sizes eta
    and returns a n-by-2 matrix of positions X"""

    nC2 = (n*(n-1))/2
    if len(d) != nC2 or len(w) != nC2:
        raise "d and w are not correct length condensed distance matrices"

    if etas is None:
        eta_max = 1/min(w)
        eta_min = .1/max(w)
        lambd = np.log(eta_max / eta_min) / 14
        etas = eta_max * np.exp(-lambd * np.arange(15))

    # initialize positions
    X = np.random.rand(n, 2)

    # do mds
    cpp.mds_direct(X, d, w, etas)
    return X



### no c++ bindings for functions below ###

def random_init(I, J):
    if len(I) != len(J):
        raise "length of edge indices I and J not equal"

    n = max(max(I), max(J)) + 1
    X = np.random.rand(n,2)
    return X


def draw_png(X, I, J, filepath, noderadius=.2, linkwidth=.05, width=1000, border=50, nodeopacity=1, linkopacity=1):
    """Takes a n-by-2 matrix of positions X and index pairs I and J
    and draws it to a .png file, using the Pillow library."""

    n = len(X)
    m = len(I)

    X_min = [min(X[i,0] for i in range(n)), min(X[i,1] for i in range(n))]
    X_max = [max(X[i,0] for i in range(n)), max(X[i,1] for i in range(n))]

    range_max = max(X_max[0]-X_min[0], X_max[1]-X_min[1]) # taller or wider
    range_max += 2*noderadius # guarantee no nodes are cut off at the edges
    scale = (width-2*border) / range_max

    X_png = np.empty((n,2))
    for i in range(n):
        X_png[i] = (X[i] - X_min) * scale
        X_png[i] += [border + scale*noderadius, border + scale*noderadius]

    # use cairo to draw
    import cairo
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, width)
    ctx = cairo.Context(surface)

    ctx.set_line_width(linkwidth * scale)
    ctx.set_line_cap(cairo.LineCap.ROUND)
    ctx.set_source_rgba(0,0,0,linkopacity)

    # draw links
    for ij in range(m):
        i = I[ij]
        j = J[ij]
        X_i = X_png[i]
        X_j = X_png[j]
        ctx.move_to(X_i[0], X_i[1])
        ctx.line_to(X_j[0], X_j[1])
        ctx.stroke()

    # draw nodes
    if noderadius > 0 and nodeopacity > 0:
        ctx.set_source_rgba(0,0,0,nodeopacity)
        radius = noderadius * scale
        for i in range(n):
            ctx.arc(X_png[i][0], X_png[i][1], radius, 0, 7)
            ctx.fill()

    surface.write_to_png(filepath)


def draw_svg(X, I, J, filepath=None, noderadius=.2, linkwidth=.05, width=1000, border=50, nodeopacity=1, linkopacity=1):
    """Takes a n-by-2 matrix of positions X and index pairs I and J
    and writes the equivalent picture in svg format.
    The drawing will be expanded into a width*width square
    Note that the parameters are in svg pixel units.
    The style at the top of the output svg may also be edited as necessary.
    The svg is returned as a string if filename is empty."""

    n = len(X)
    m = len(I)

    X_min = [min(X[i,0] for i in range(n)), min(X[i,1] for i in range(n))]
    X_max = [max(X[i,0] for i in range(n)), max(X[i,1] for i in range(n))]

    range_max = max(X_max[0]-X_min[0], X_max[1]-X_min[1]) # taller or wider
    range_max += 2*noderadius # guarantee no nodes are cut off at the edges
    scale = (width-2*border) / range_max

    X_svg = np.empty((n,2))
    for i in range(n):
        X_svg[i] = (X[i] - X_min) * scale
        X_svg[i] += [border + scale*noderadius, border + scale*noderadius]

    svg_list = []
    svg_list.append('<svg width="{:.0f}" height="{:.0f}" xmlns="http://www.w3.org/2000/svg">'.format(width, width))
    svg_list.append('<style type="text/css">')
    svg_list.append('line{{stroke:black;stroke-width:{:.3f};stroke-opacity:{:.3f};stroke-linecap:round;}}'.format(scale*linkwidth,linkopacity))
    svg_list.append('circle{{r:{};fill:black;fill-opacity:{:.3f}}}'.format(scale*noderadius,nodeopacity))
    svg_list.append('</style>');

    # draw links
    for ij in range(m):
        i = I[ij]
        j = J[ij]
        X_i = X_svg[i]
        X_j = X_svg[j]
        svg_list.append('<line x1="{:.1f}" x2="{:.1f}" y1="{:.1f}" y2="{:.1f}"/>'.format(X_i[0], X_j[0], X_i[1], X_j[1]));

    # draw nodes
    if noderadius > 0:
        for i in range(n):
            svg_list.append('<circle cx="{:.1f}" cy="{:.1f}"/>'.format(X_svg[i][0], X_svg[i][1]))

    svg_list.append("</svg>")

    if filepath == None or filepath == '':
        return '\n'.join(svg_list)
    else:
        f = open(filepath, 'w')
        f.write('\n'.join(svg_list))
        f.close()



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.8
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_layout', [dirname(__file__)])
        except ImportError:
            import _layout
            return _layout
        if fp is not None:
            try:
                _mod = imp.load_module('_layout', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _layout = swig_import_helper()
    del swig_import_helper
else:
    import _layout
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0



def dimension_check(kd):
    return _layout.dimension_check(kd)
dimension_check = _layout.dimension_check

def unweighted_edge_check(len_I, len_J):
    return _layout.unweighted_edge_check(len_I, len_J)
unweighted_edge_check = _layout.unweighted_edge_check

def weighted_edge_check(len_I, len_J, len_V):
    return _layout.weighted_edge_check(len_I, len_J, len_V)
weighted_edge_check = _layout.weighted_edge_check

def layout_unweighted(X, I, J, t_max, eps):
    return _layout.layout_unweighted(X, I, J, t_max, eps)
layout_unweighted = _layout.layout_unweighted

def layout_weighted(X, I, J, V, t_max, eps):
    return _layout.layout_weighted(X, I, J, V, t_max, eps)
layout_weighted = _layout.layout_weighted

def layout_unweighted_convergent(X, I, J, t_max, eps, delta, t_maxmax):
    return _layout.layout_unweighted_convergent(X, I, J, t_max, eps, delta, t_maxmax)
layout_unweighted_convergent = _layout.layout_unweighted_convergent

def layout_weighted_convergent(X, I, J, V, t_max, eps, delta, t_maxmax):
    return _layout.layout_weighted_convergent(X, I, J, V, t_max, eps, delta, t_maxmax)
layout_weighted_convergent = _layout.layout_weighted_convergent

def layout_sparse_unweighted(X, I, J, p, t_max, eps):
    return _layout.layout_sparse_unweighted(X, I, J, p, t_max, eps)
layout_sparse_unweighted = _layout.layout_sparse_unweighted

def layout_sparse_weighted(X, I, J, V, p, t_max, eps):
    return _layout.layout_sparse_weighted(X, I, J, V, p, t_max, eps)
layout_sparse_weighted = _layout.layout_sparse_weighted

def mds_direct(X, d, w, eta):
    return _layout.mds_direct(X, d, w, eta)
mds_direct = _layout.mds_direct
# This file is compatible with both classic and new-style classes.





from setuptools import setup
from setuptools.extension import Extension

# Third-party modules - we depend on numpy for everything
import numpy
# Obtain the numpy include directory.  This logic works across numpy versions.
try:
    numpy_include = numpy.get_include()
except AttributeError:
    numpy_include = numpy.get_numpy_include()

_layout = Extension(
    name="_layout",
    sources=["layout.cpp", "sparse.cpp", "layout_wrap.cxx"],
    extra_compile_args=["-std=c++11"],
    include_dirs=[numpy_include]
)

setup(
    name="s_gd2",
    version="1.2",
    author="Jonathan Zheng",
    author_email="jxz12@ic.ac.uk",
    url="https://www.github.com/jxz12/s_gd2",
    description="A package for performing stochastic gradient descent (arXiv:1710.04626) to layout graphs",
    install_requires=['numpy', 'pycairo'],
    py_modules=['s_gd2', 'layout'],
    ext_modules=[_layout]
)

